// file: Binding.java
// author: CS102 S14 Class
// date: March 20, 2014
//
// This is an API for an immutable map. Unlike java.util.Map and ST in SW 3.1, 
// the get function for this map returns an Option type to deal with the case 
// where a key was not in the map.
//
public class Binding<Key, Value> implements Map<Key, Value> {
 
    private Key key;
    private Value value;
    private Map<Key, Value> next;

    private int N;
    
    public Binding(Key key, Value value, Map<Key, Value> next) {
        this.key = key;
        this.value = value;
        this.next = next;
        this.N = next.size() + 1;
    }
    
    // Getters, there are no setters.(!)
    //
    private Key getKey()     { return this.key; }
    private Value getValue() { return this.value; }
    private Map<Key, Value> getNext() { return this.next; }
    
    public Option<Value> get(Key key) {
        if (this.getKey().equals(key))
            return new Some(this.getValue());
        else
            return this.getNext().get(key);
    }
    
    public Map<Key, Value> put(Key key, Value value) {
        return new Binding(key, value, this);
    }
    
    public Map<Key, Value> remove(Key key) {
        Map<Key, Value> nextAnswer = this.getNext().remove(key);
        if (this.getKey().equals(key))
            return nextAnswer;
        else
            return new Binding(this.getKey(), this.getValue(), nextAnswer);
    }
    
    public int size()        { return this.N; }
    public boolean isEmpty() { return this.size() == 0; }
    
    public String toString() {
        StringBuilder sb = new StringBuilder();
        String 
            ks = this.getKey().toString(),
            vs = this.getValue().toString(),
            rs = this.getNext().toString();
        return "[" + ks + "=" + vs + " " + rs + "]";
    }
    
    public static void main(String[] args) {        
        Map<String, String> m0 = new Empty(),
            m1 = m0.put("A", "Alice"),
            m2 = m1.put("B", "Bob"),
            m3 = m2.put("A", "Tommy"),
            m4 = m3.remove("A");
        System.out.println("map = " + m4.toString());
    }
}